@@ -24,9 +24,6 @@
 #include <string.h>
 #include <time.h>
 #include <unistd.h>
-#ifdef HAVE_GETOPT_H
-# include <getopt.h>
-#endif
 #include <stdlib.h>
 #include <npth.h>
 #ifdef HAVE_W32_SYSTEM
@@ -54,8 +51,7 @@
 #define SERVERINFO_PGPKEYV2 2 /* Needs "pgpKeyV2" instead of "pgpKey"*/
 #define SERVERINFO_SCHEMAV2 4 /* Version 2 of the Schema.            */
 #define SERVERINFO_NTDS     8 /* Server is an Active Directory.      */
-#define SERVERINFO_GENERIC 16 /* Connected in generic mode.          */
-#define SERVERINFO_CNFPR   32 /* Server uses CN=fingerprint.         */
+#define SERVERINFO_GENERIC 16 /* Connected in genric mode.           */
 
 
 /* The page size requested from the server.  */
@@ -303,16 +299,15 @@ ks_ldap_help (ctrl_t ctrl, parsed_uri_t uri)
 
   if(!uri)
     err = ks_print_help (ctrl, "  ldap");
-  else if (!strcmp (uri->scheme, "ldap")
-           || !strcmp (uri->scheme, "ldaps")
-           || !strcmp (uri->scheme, "ldapi")
-           || uri->opaque)
+  else if (uri->is_ldap || uri->opaque)
     err = ks_print_help (ctrl, data);
   else
     err = 0;
 
   return err;
 }
+
+
 
 /* Create a new empty state object.  Returns NULL on error */
 static struct ks_engine_ldap_local_s *
@@ -494,12 +489,10 @@ keyspec_to_ldap_filter (const char *keyspec, char **filter, int only_exact,
 		     (ulong) desc.u.kid[0], (ulong) desc.u.kid[1]);
       break;
 
-    case KEYDB_SEARCH_MODE_FPR16:
-    case KEYDB_SEARCH_MODE_FPR20:
     case KEYDB_SEARCH_MODE_FPR:
       if ((serverinfo & SERVERINFO_SCHEMAV2))
         {
-          freeme = bin2hex (desc.u.fpr, 20, NULL);
+          freeme = bin2hex (desc.u.fpr, desc.fprlen, NULL);
           if (!freeme)
             return gpg_error_from_syserror ();
           f = xasprintf ("(|(gpgFingerprint=%s)(gpgSubFingerprint=%s))",
@@ -583,17 +576,14 @@ interrogate_ldap_dn (LDAP *ldap_conn, const char *basedn_search,
             log_debug ("Version:\t%s\n", vals[0]);
           if (is_gnupg)
             {
-              char *fields[2];
+              const char *fields[2];
               int nfields;
               nfields = split_fields (vals[0], fields, DIM(fields));
               if (nfields > 0 && atoi(fields[0]) > 1)
                 *r_serverinfo |= SERVERINFO_SCHEMAV2;
               if (nfields > 1
                   && !ascii_strcasecmp (fields[1], "ntds"))
-                *r_serverinfo |= (SERVERINFO_NTDS|SERVERINFO_CNFPR);
-              else if (nfields > 1
-                       && !ascii_strcasecmp (fields[1], "cnfpr"))
-                *r_serverinfo |= SERVERINFO_CNFPR;
+                *r_serverinfo |= SERVERINFO_NTDS;
             }
         }
       my_ldap_value_free (vals);
@@ -615,7 +605,7 @@ interrogate_ldap_dn (LDAP *ldap_conn, const char *basedn_search,
  * including whether to use TLS and the username and password (see
  * ldap_parse_uri for a description of the various fields).  Be
  * default a PGP keyserver is assumed; if GENERIC is true a generic
- * ldap conenction is instead established.
+ * ldap connection is instead established.
  *
  * Returns: The ldap connection handle in *LDAP_CONNP, R_BASEDN is set
  * to the base DN for the PGP key space, several flags will be stored
@@ -865,11 +855,7 @@ my_ldap_connect (parsed_uri_t uri, unsigned int generic, LDAP **ldap_connp,
   else if (bindname)
     {
       npth_unprotect ();
-      /* Older Windows header dont have the const for the last two args.
-       * Thus we need to cast to avoid warnings.  */
-      lerr = ldap_simple_bind_s (ldap_conn,
-                                 (char * const)bindname,
-                                 (char * const)password);
+      lerr = ldap_simple_bind_s (ldap_conn, bindname, password);
       npth_protect ();
       if (lerr != LDAP_SUCCESS)
 	{
@@ -1158,7 +1144,7 @@ no_ldap_due_to_tor (ctrl_t ctrl)
 
   log_error ("%s", msg);
   dirmngr_status_printf (ctrl, "NOTE", "no_ldap_due_to_tor %u %s", err, msg);
-  return err;
+  return gpg_error (GPG_ERR_NOT_SUPPORTED);
 }
 
 
@@ -1645,9 +1631,10 @@ ks_ldap_get (ctrl_t ctrl, parsed_uri_t uri, const char *keyspec,
      NULL
     };
 
-
   if (dirmngr_use_tor ())
-    return no_ldap_due_to_tor (ctrl);
+    {
+      return no_ldap_due_to_tor (ctrl);
+    }
 
   err = ks_ldap_prepare_my_state (ctrl, ks_get_flags, &first_mode, &next_mode);
   if (err)
@@ -1875,7 +1862,9 @@ ks_ldap_search (ctrl_t ctrl, parsed_uri_t uri, const char *pattern,
   (void) ctrl;
 
   if (dirmngr_use_tor ())
-    return no_ldap_due_to_tor (ctrl);
+    {
+      return no_ldap_due_to_tor (ctrl);
+    }
 
   /* Make sure we are talking to an OpenPGP LDAP server.  */
   err = my_ldap_connect (uri, 0, &ldap_conn, &basedn, NULL, NULL, &serverinfo);
@@ -2716,7 +2705,7 @@ extract_attributes (LDAPMod ***modlist, int *extract_state,
 
       uncescape (uid);
       modlist_add (modlist, "pgpUserID", uid);
-      if (schemav2 && (mbox = mailbox_from_userid (uid)))
+      if (schemav2 && (mbox = mailbox_from_userid (uid, 0)))
         {
           modlist_add (modlist, "gpgMailbox", mbox);
           xfree (mbox);
@@ -2733,7 +2722,7 @@ extract_attributes (LDAPMod ***modlist, int *extract_state,
 gpg_error_t
 ks_ldap_put (ctrl_t ctrl, parsed_uri_t uri,
 	     void *data, size_t datalen,
-	     void *info_arg, size_t infolen_arg)
+	     void *info, size_t infolen)
 {
   gpg_error_t err = 0;
   int ldap_err;
@@ -2744,8 +2733,9 @@ ks_ldap_put (ctrl_t ctrl, parsed_uri_t uri,
   LDAPMod **addlist = NULL;
   char *data_armored = NULL;
   int extract_state;
-  char *info = info_arg;
-  char *info_buffer = NULL;  /* Used in case we need to copy INFO.  */
+
+  /* The last byte of the info block.  */
+  const char *infoend = (const char *) info + infolen - 1;
 
   /* Enable this code to dump the modlist to /tmp/modlist.txt.  */
 #if 0
@@ -2756,30 +2746,14 @@ ks_ldap_put (ctrl_t ctrl, parsed_uri_t uri,
 #endif
   estream_t dump = NULL;
 
+  /* Elide a warning.  */
+  (void) ctrl;
+
   if (dirmngr_use_tor ())
-    return no_ldap_due_to_tor (ctrl);
-
-  /* Make sure that INFO is a proper string. */
-  if (!infolen_arg)
     {
-      /* Too short - use an empty string.  */
-      info = info_buffer = xtrystrdup ("");
-    }
-  else if (info[infolen_arg-1])
-    {
-      /* No terminating nul.  Take a copy.  */
-      info_buffer = xtrymalloc (infolen_arg + 1);
-      if (!info_buffer)
-        {
-          err = gpg_error_from_syserror ();
-          goto out;
-        }
-      memcpy (info_buffer, info, infolen_arg);
-      info_buffer[infolen_arg] = '\0';
-      info = info_buffer;
+      return no_ldap_due_to_tor (ctrl);
     }
 
-  /* Connect to the LDAP server to get its properties.  */
   err = my_ldap_connect (uri, 0, &ldap_conn, &basedn, NULL, NULL, &serverinfo);
   if (err || !basedn)
     {
@@ -2843,8 +2817,8 @@ ks_ldap_put (ctrl_t ctrl, parsed_uri_t uri,
       if (dump)
 	{
 	  es_fprintf(dump, "data (%zd bytes)\n", datalen);
-	  es_fprintf(dump, "info (%zd bytes): '\n", strlen (info));
-	  es_fwrite(info, strlen (info), 1, dump);
+	  es_fprintf(dump, "info (%zd bytes): '\n", infolen);
+	  es_fwrite(info, infolen, 1, dump);
 	  es_fprintf(dump, "'\n");
 	}
     }
@@ -2871,24 +2845,37 @@ ks_ldap_put (ctrl_t ctrl, parsed_uri_t uri,
 
   /* Assemble the INFO stuff into LDAP attributes */
   extract_state = 0;
-  while (*info)
+  while (infolen > 0)
     {
-      char *newline;
+      char *temp = NULL;
 
-      newline = strchr (info, '\n');
-      if (newline)
-        *newline = '\0';  /* Replace the LF */
+      char *newline = memchr (info, '\n', infolen);
+      if (! newline)
+	/* The last line is not \n terminated!  Make a copy so we can
+	   add a NUL terminator.  */
+	{
+	  temp = xmalloc (infolen + 1);
+	  memcpy (temp, info, infolen);
+	  info = temp;
+	  newline = (char *) info + infolen;
+	}
+
+      *newline = '\0';
 
       extract_attributes (&addlist, &extract_state, info,
                           (serverinfo & SERVERINFO_SCHEMAV2));
 
-      if (newline)
-        {
-          *newline = '\n';     /* Restore the LF  */
-          info = newline + 1;  /* and forward.    */
-        }
+      infolen = infolen - ((uintptr_t) newline - (uintptr_t) info + 1);
+      info = newline + 1;
+
+      /* Sanity check.  */
+      if (! temp)
+	log_assert ((char *) info + infolen - 1 == infoend);
       else
-        info += strlen (info); /* Forward to the Nul.  */
+	{
+	  log_assert (infolen == -1);
+	  xfree (temp);
+	}
     }
 
   modlist_add (&addlist, "objectClass", "pgpKeyInfo");
@@ -2925,7 +2912,7 @@ ks_ldap_put (ctrl_t ctrl, parsed_uri_t uri,
     char **attrval;
     char *dn;
 
-    if ((serverinfo & SERVERINFO_CNFPR))
+    if ((serverinfo & SERVERINFO_NTDS))
       {
         /* The modern way using a CN RDN with the fingerprint.  This
          * has the advantage that we won't have duplicate 64 bit
@@ -2990,117 +2977,7 @@ ks_ldap_put (ctrl_t ctrl, parsed_uri_t uri,
   xfree (addlist);
 
   xfree (data_armored);
-  xfree (info_buffer);
-  return err;
-}
 
-
-/* Delete the keys given by the list of fingerprints in FPRLIST from
- * the keyserver identified by URI.  The function stops at the first
- * error encountered. */
-gpg_error_t
-ks_ldap_del (ctrl_t ctrl, parsed_uri_t uri, strlist_t fprlist)
-{
-  gpg_error_t err = 0;
-  int ldap_err;
-  unsigned int serverinfo;
-  LDAP *ldap_conn = NULL;
-  char *basedn = NULL;
-  char *dn = NULL;
-  strlist_t fpr;
-  unsigned int count = 0;
-  unsigned int totalcount = 0;
-
-  if (dirmngr_use_tor ())
-    {
-      return no_ldap_due_to_tor (ctrl);
-    }
-
-  for (fpr = fprlist; fpr; fpr = fpr->next)
-    totalcount++;
-
-  err = my_ldap_connect (uri, 0, &ldap_conn, &basedn, NULL, NULL, &serverinfo);
-  if (err || !basedn)
-    {
-      if(opt.verbose)
-        log_info ("%s: connecting to server failed\n", __func__);
-      if (!err)
-	err = gpg_error (GPG_ERR_GENERAL);  /* (no baseDN) */
-      goto leave;
-    }
-
-  if (!(serverinfo & SERVERINFO_REALLDAP))
-    {
-      if(opt.verbose)
-        log_info ("%s: The PGP.com keyserver is not supported\n", __func__);
-      err = gpg_error (GPG_ERR_NOT_SUPPORTED);
-      goto leave;
-    }
-
-  if (!(serverinfo & SERVERINFO_SCHEMAV2))
-    {
-      if(opt.verbose)
-        log_info ("%s: The keyserver does not support the v2 schema\n",
-                  __func__);
-      err = gpg_error (GPG_ERR_NOT_SUPPORTED);
-      goto leave;
-    }
-
-  if (opt.verbose)
-    log_info ("%s: Using DN: %s,%s\n", __func__,
-              (serverinfo & SERVERINFO_CNFPR)? "CN=<fingerprint>"
-              /* */                          : "pgpCertID=<keyid>",
-              basedn);
-  for (fpr = fprlist; fpr; fpr = fpr->next)
-    {
-      if ((serverinfo & SERVERINFO_CNFPR))
-        {
-          xfree (dn);
-          dn = xtryasprintf ("CN=%s,%s", fpr->d, basedn);
-        }
-      else
-        {
-          unsigned int off;
-
-          /* Simle method to get the keyID.  Note that a v5 key
-           * (len>40) has the keyid at the left.  If the length is
-           * less than 17 we assume a keyid has been given.  */
-          off = strlen (fpr->d);
-          if (off <= 40 && off > 16)
-            off = off - 16;
-          else
-            off = 0;
-
-          xfree (dn);
-          dn = xtryasprintf ("pgpCertID=%.16s,%s", fpr->d+off, basedn);
-        }
-
-      npth_unprotect ();
-      ldap_err = ldap_delete_ext_s (ldap_conn, dn, NULL, NULL);
-      npth_protect ();
-      if (ldap_err == LDAP_SUCCESS)
-        {
-          if (opt.verbose)
-            log_info ("%s: key %s deleted\n", __func__, fpr->d);
-          count++;
-        }
-      else
-        {
-          log_error ("%s: error deleting key %s: %s\n",
-                     __func__, fpr->d, ldap_err2string (ldap_err));
-          err = ldap_err_to_gpg_err (ldap_err);
-          break;  /* Stop at the first failed deletion.  */
-        }
-    }
-  log_info ("%s: number of keys deleted: %u of %u\n",
-            __func__, count, totalcount);
-
-
- leave:
-  if (ldap_conn)
-    ldap_unbind (ldap_conn);
-  xfree (dn);
-  xfree (basedn);
   return err;
 }
 
