@@ -25,7 +25,6 @@
 #include <errno.h>
 #include <unistd.h>
 #include <time.h>
-#include <assert.h>
 #ifdef HAVE_LOCALE_H
 #include <locale.h>
 #endif
@@ -76,123 +75,31 @@ struct import_key_parm_s
   size_t keylen;
 };
 
+struct sethash_inq_parm_s
+{
+  assuan_context_t ctx;
+  const void *data;
+  size_t datalen;
+};
+
 struct default_inq_parm_s
 {
   ctrl_t ctrl;
   assuan_context_t ctx;
 };
 
-
-/* An object and variable to cache ISTRUSTED calls.  The cache is
- * global and reset with each mark trusted.  We also have a disabled
- * flag here in case the gpg-agent did not allow us to query all
- * trusted keys at once.  */
-struct istrusted_cache_s
-{
-  struct istrusted_cache_s *next;
-  struct rootca_flags_s flags; /* The flags of this fingerprint.  */
-  char fpr[1];  /* The fingerprint of the trusted key in hex format.  */
-};
-typedef struct istrusted_cache_s *istrusted_cache_t;
-static istrusted_cache_t istrusted_cache;
-static int istrusted_cache_valid;
-static int istrusted_cache_disabled;
-
-/* Flag indicating that we can't use the keyinfo cache at all.  The
- * actual cache is stored in CTRL.  */
-static int keyinfo_cache_disabled;
-
-
 
-static void
-flush_istrusted_cache (void)
-{
-  istrusted_cache_t mycache;
-
-  /* First unlink the cache to be npth safe.  Note that we don't clear
-   * the the disabled flag - this is considered a permantent error. */
-  mycache = istrusted_cache;
-  istrusted_cache = NULL;
-  istrusted_cache_valid = 0;
-
-  while (mycache)
-    {
-      istrusted_cache_t next = mycache->next;
-      xfree (mycache);
-      mycache = next;
-    }
-}
-
-
-/* Release all items in *CACHEP and set CACHEP to NULL  */
-static void
-release_a_keyinfo_cache (keyinfo_cache_item_t *cachep)
-{
-  keyinfo_cache_item_t mycache;
-
-  /* First unlink the cache to be npth safe.  */
-  mycache = *cachep;
-  *cachep = NULL;
-
-  while (mycache)
-    {
-      keyinfo_cache_item_t next = mycache->next;
-      xfree (mycache);
-      mycache = next;
-    }
-}
-
-
-/* Flush the keyinfo cache for the session CTRL.  */
-void
-gpgsm_flush_keyinfo_cache (ctrl_t ctrl)
-{
-  ctrl->keyinfo_cache_valid = 0;
-  release_a_keyinfo_cache (&ctrl->keyinfo_cache);
-}
-
 /* Print a warning if the server's version number is less than our
    version number.  Returns an error code on a connection problem.  */
 static gpg_error_t
 warn_version_mismatch (ctrl_t ctrl, assuan_context_t ctx,
                        const char *servername, int mode)
 {
-  gpg_error_t err;
-  char *serverversion;
-  const char *myversion = strusage (13);
-
-  err = get_assuan_server_version (ctx, mode, &serverversion);
-  if (err)
-    log_error (_("error getting version from '%s': %s\n"),
-               servername, gpg_strerror (err));
-  else if (compare_version_strings (serverversion, myversion) < 0)
-    {
-      char *warn;
-
-      warn = xtryasprintf (_("server '%s' is older than us (%s < %s)"),
-                           servername, serverversion, myversion);
-      if (!warn)
-        err = gpg_error_from_syserror ();
-      else
-        {
-          log_info (_("WARNING: %s\n"), warn);
-          if (!opt.quiet)
-            {
-              log_info (_("Note: Outdated servers may lack important"
-                          " security fixes.\n"));
-              log_info (_("Note: Use the command \"%s\" to restart them.\n"),
-                        "gpgconf --kill all");
-            }
-          gpgsm_status2 (ctrl, STATUS_WARNING, "server_version_mismatch 0",
-                         warn, NULL);
-          xfree (warn);
-        }
-    }
-  xfree (serverversion);
-  return err;
+  return warn_server_version_mismatch (ctx, servername, mode,
+                                       gpgsm_status2, ctrl,
+                                       !opt.quiet);
 }
 
-
 /* Try to connect to the agent via socket or fork it off and work by
    pipes.  Handle the server's initial greeting */
 static int
@@ -327,8 +234,29 @@ default_inq_cb (void *opaque, const char *line)
 
 
 
+/* This is the inquiry callback required by the SETHASH command.  */
+static gpg_error_t
+sethash_inq_cb (void *opaque, const char *line)
+{
+  gpg_error_t err = 0;
+  struct sethash_inq_parm_s *parm = opaque;
+
+  if (has_leading_keyword (line, "TBSDATA"))
+    {
+      err = assuan_send_data (parm->ctx, parm->data, parm->datalen);
+    }
+  else
+    log_error ("ignoring gpg-agent inquiry '%s'\n", line);
+
+  return err;
+}
+
+
 /* Call the agent to do a sign operation using the key identified by
-   the hex string KEYGRIP. */
+ * the hex string KEYGRIP.  If DIGESTALGO is given (DIGEST,DIGESTLEN)
+ * gives the to be signed hash created using the given algo.  If
+ * DIGESTALGO is not given (i.e. zero) (DIGEST,DIGESTALGO) give the
+ * entire data to-be-signed. */
 int
 gpgsm_agent_pksign (ctrl_t ctrl, const char *keygrip, const char *desc,
                     unsigned char *digest, size_t digestlen, int digestalgo,
@@ -347,7 +275,7 @@ gpgsm_agent_pksign (ctrl_t ctrl, const char *keygrip, const char *desc,
   inq_parm.ctrl = ctrl;
   inq_parm.ctx = agent_ctx;
 
-  if (digestlen*2 + 50 > DIM(line))
+  if (digestalgo && digestlen*2 + 50 > DIM(line))
     return gpg_error (GPG_ERR_GENERAL);
 
   rc = assuan_transact (agent_ctx, "RESET", NULL, NULL, NULL, NULL, NULL, NULL);
@@ -368,11 +296,26 @@ gpgsm_agent_pksign (ctrl_t ctrl, const char *keygrip, const char *desc,
         return rc;
     }
 
-  sprintf (line, "SETHASH %d ", digestalgo);
-  p = line + strlen (line);
-  for (i=0; i < digestlen ; i++, p += 2 )
-    sprintf (p, "%02X", digest[i]);
-  rc = assuan_transact (agent_ctx, line, NULL, NULL, NULL, NULL, NULL, NULL);
+  if (!digestalgo)
+    {
+      struct sethash_inq_parm_s sethash_inq_parm;
+
+      sethash_inq_parm.ctx = agent_ctx;
+      sethash_inq_parm.data = digest;
+      sethash_inq_parm.datalen = digestlen;
+      rc = assuan_transact (agent_ctx, "SETHASH --inquire",
+                            NULL, NULL, sethash_inq_cb, &sethash_inq_parm,
+                            NULL, NULL);
+    }
+  else
+    {
+      snprintf (line, sizeof line, "SETHASH %d ", digestalgo);
+      p = line + strlen (line);
+      for (i=0; i < digestlen ; i++, p += 2 )
+        sprintf (p, "%02X", digest[i]);
+      rc = assuan_transact (agent_ctx, line,
+                            NULL, NULL, NULL, NULL, NULL, NULL);
+    }
   if (rc)
     return rc;
 
@@ -424,7 +367,6 @@ gpgsm_scd_pksign (ctrl_t ctrl, const char *keyid, const char *desc,
     case GCRY_MD_RMD160:hashopt = "--hash=rmd160"; break;
     case GCRY_MD_MD5:   hashopt = "--hash=md5"; break;
     case GCRY_MD_SHA256:hashopt = "--hash=sha256"; break;
-    case GCRY_MD_SHA384:hashopt = "--hash=sha384"; break;
     case GCRY_MD_SHA512:hashopt = "--hash=sha512"; break;
     default:
       return gpg_error (GPG_ERR_DIGEST_ALGO);
@@ -480,13 +422,19 @@ gpgsm_scd_pksign (ctrl_t ctrl, const char *keyid, const char *desc,
     {
     case GCRY_PK_RSA:
       rc = gcry_sexp_build (&sig, NULL, "(sig-val(rsa(s%b)))",
-                            sigbuflen, sigbuf);
+                            (int)sigbuflen, sigbuf);
       break;
 
     case GCRY_PK_ECC:
       rc = gcry_sexp_build (&sig, NULL, "(sig-val(ecdsa(r%b)(s%b)))",
-                            sigbuflen/2, sigbuf,
-                            sigbuflen/2, sigbuf + sigbuflen/2);
+                            (int)sigbuflen/2, sigbuf,
+                            (int)sigbuflen/2, sigbuf + sigbuflen/2);
+      break;
+
+    case GCRY_PK_EDDSA:
+      rc = gcry_sexp_build (&sig, NULL, "(sig-val(eddsa(r%b)(s%b)))",
+                            (int)sigbuflen/2, sigbuf,
+                            (int)sigbuflen/2, sigbuf + sigbuflen/2);
       break;
 
     default:
@@ -502,7 +450,7 @@ gpgsm_scd_pksign (ctrl_t ctrl, const char *keyid, const char *desc,
   if (rc)
     return rc;
 
-  assert (gcry_sexp_canon_len (*r_buf, *r_buflen, NULL, NULL));
+  log_assert (gcry_sexp_canon_len (*r_buf, *r_buflen, NULL, NULL));
   return  0;
 }
 
@@ -564,7 +512,7 @@ gpgsm_agent_pkdecrypt (ctrl_t ctrl, const char *keygrip, const char *desc,
   if (rc)
     return rc;
 
-  assert ( DIM(line) >= 50 );
+  log_assert ( DIM(line) >= 50 );
   snprintf (line, DIM(line), "SETKEY %s", keygrip);
   rc = assuan_transact (agent_ctx, line, NULL, NULL, NULL, NULL, NULL, NULL);
   if (rc)
@@ -598,7 +546,7 @@ gpgsm_agent_pkdecrypt (ctrl_t ctrl, const char *keygrip, const char *desc,
   buf = get_membuf (&data, &len);
   if (!buf)
     return gpg_error (GPG_ERR_ENOMEM);
-  assert (len); /* (we forced Nul termination.)  */
+  log_assert (len); /* (we forced Nul termination.)  */
 
   if (*buf == '(')
     {
@@ -929,61 +877,29 @@ gpgsm_agent_scd_keypairinfo (ctrl_t ctrl, strlist_t *r_list)
 
 
 
-struct istrusted_status_parm_s
-{
-  struct rootca_flags_s flags;
-  istrusted_cache_t cache;
-};
-
-
 static gpg_error_t
 istrusted_status_cb (void *opaque, const char *line)
 {
-  struct istrusted_status_parm_s *parm = opaque;
+  struct rootca_flags_s *flags = opaque;
   const char *s;
 
   if ((s = has_leading_keyword (line, "TRUSTLISTFLAG")))
     {
       line = s;
       if (has_leading_keyword (line, "relax"))
-        parm->flags.relax = 1;
+        flags->relax = 1;
       else if (has_leading_keyword (line, "cm"))
-        parm->flags.chain_model = 1;
+        flags->chain_model = 1;
       else if (has_leading_keyword (line, "qual"))
-        parm->flags.qualified = 1;
+        flags->qualified = 1;
       else if (has_leading_keyword (line, "de-vs"))
-        parm->flags.de_vs = 1;
-
-      /* Copy the current flags to the current list item.  */
-      if (parm->cache)
-        parm->cache->flags = parm->flags;
-    }
-  else if ((s = has_leading_keyword (line, "TRUSTLISTFPR")) && *s)
-    {
-      /* We see this only with the "LISTTRUSTED --status" command but
-       * not with ISTRUSTED.  Thus the cache will only be filled by
-       * the former command. */
-      istrusted_cache_t ci;
-
-      ci = xtrymalloc (sizeof *ci + strlen (s));
-      if (!ci)
-        return gpg_error_from_syserror ();
-      strcpy (ci->fpr, s);
-      memset (&ci->flags, 0, sizeof ci->flags);
-
-      if ((opt.compat_flags & COMPAT_DE_VS_TRUSTLIST))
-        parm->flags.de_vs = 1;
-
-      ci->next = parm->cache;
-      parm->cache = ci;
-      /* Also need to clear the parm's flags which will be copied to
-       * the cache.  */
-      memset (&parm->flags, 0, sizeof ci->flags);
+        flags->de_vs = 1;
     }
   return 0;
 }
 
 
+
 /* Ask the agent whether the certificate is in the list of trusted
    keys.  The certificate is either specified by the CERT object or by
    the fingerprint HEXFPR.  ROOTCA_FLAGS is guaranteed to be cleared
@@ -994,12 +910,8 @@ gpgsm_agent_istrusted (ctrl_t ctrl, ksba_cert_t cert, const char *hexfpr,
 {
   int rc;
   char line[ASSUAN_LINELENGTH];
-  char *fpr_buffer = NULL;
-  struct istrusted_status_parm_s parm;
-  istrusted_cache_t ci;
 
   memset (rootca_flags, 0, sizeof *rootca_flags);
-  memset (&parm, 0, sizeof parm);
 
   if (cert && hexfpr)
     return gpg_error (GPG_ERR_INV_ARG);
@@ -1008,70 +920,29 @@ gpgsm_agent_istrusted (ctrl_t ctrl, ksba_cert_t cert, const char *hexfpr,
   if (rc)
     return rc;
 
-  if (!hexfpr)
+  if (hexfpr)
     {
-      fpr_buffer = gpgsm_get_fingerprint_hexstring (cert, GCRY_MD_SHA1);
-      if (!fpr_buffer)
+      snprintf (line, DIM(line), "ISTRUSTED %s", hexfpr);
+    }
+  else
+    {
+      char *fpr;
+
+      fpr = gpgsm_get_fingerprint_hexstring (cert, GCRY_MD_SHA1);
+      if (!fpr)
         {
           log_error ("error getting the fingerprint\n");
-          rc = gpg_error (GPG_ERR_GENERAL);
-          goto leave;
+          return gpg_error (GPG_ERR_GENERAL);
         }
-      hexfpr = fpr_buffer;
+
+      snprintf (line, DIM(line), "ISTRUSTED %s", fpr);
+      xfree (fpr);
     }
 
-  /* First try to get the info from the cache.  */
-  if ((opt.compat_flags & COMPAT_NO_KEYINFO_CACHE))
-    istrusted_cache_disabled = 1;
-
-  if (!istrusted_cache_disabled && !istrusted_cache_valid)
-    {
-      /* Cache is empty - fill it.  */
-      rc = assuan_transact (agent_ctx, "LISTTRUSTED --status",
-                            NULL, NULL, NULL, NULL,
-                            istrusted_status_cb, &parm);
-      istrusted_cache = parm.cache;
-      parm.cache = NULL;
-      if (rc)
-        {
-          if (gpg_err_code (rc) != GPG_ERR_FORBIDDEN)
-            log_info ("filling istrusted cache failed: %s\n",
-                       gpg_strerror (rc));
-          istrusted_cache_disabled = 1;
-          flush_istrusted_cache ();
-          rc = 0;  /* Fallback to single requests.  */
-        }
-      else
-        istrusted_cache_valid = 1;
-    }
-
-  if (istrusted_cache_valid)
-    {
-      for (ci = istrusted_cache; ci; ci = ci->next)
-        if (!strcmp (ci->fpr, hexfpr))
-          break;  /* Found.  */
-      if (ci)
-        {
-          *rootca_flags = ci->flags;
-          rootca_flags->valid = 1;
-          rc = 0;
-        }
-      else
-        rc = gpg_error (GPG_ERR_NOT_TRUSTED);
-      goto leave;
-    }
-
-  snprintf (line, DIM(line), "ISTRUSTED %s", hexfpr);
   rc = assuan_transact (agent_ctx, line, NULL, NULL, NULL, NULL,
-                        istrusted_status_cb, &parm);
+                        istrusted_status_cb, rootca_flags);
   if (!rc)
-    {
-      *rootca_flags = parm.flags;
-      rootca_flags->valid = 1;
-    }
-
- leave:
-  xfree (fpr_buffer);
+    rootca_flags->valid = 1;
   return rc;
 }
 
@@ -1113,10 +984,6 @@ gpgsm_agent_marktrusted (ctrl_t ctrl, ksba_cert_t cert)
 
   rc = assuan_transact (agent_ctx, line, NULL, NULL,
                         default_inq_cb, &inq_parm, NULL, NULL);
-  /* Marktrusted changes the trustlist and thus we need to flush the
-   * cache.   */
-  if (!rc)
-    flush_istrusted_cache ();
   return rc;
 }
 
@@ -1376,80 +1243,42 @@ gpgsm_agent_send_nop (ctrl_t ctrl)
 
 
 
-struct keyinfo_status_parm_s
-{
-  char *serialno;
-  int fill_mode;  /* True if we want to fill the cache.  */
-  keyinfo_cache_item_t cache;
-};
-
 static gpg_error_t
 keyinfo_status_cb (void *opaque, const char *line)
 {
-  struct keyinfo_status_parm_s *parm = opaque;
-  const char *s0, *s, *s2;
+  char **serialno = opaque;
+  const char *s, *s2;
 
-  if ((s0 = has_leading_keyword (line, "KEYINFO"))
-      && (!parm->serialno || parm->fill_mode))
+  if ((s = has_leading_keyword (line, "KEYINFO")) && !*serialno)
     {
-      s = strchr (s0, ' ');
-      xfree (parm->serialno);
-      parm->serialno = NULL;
+      s = strchr (s, ' ');
       if (s && s[1] == 'T' && s[2] == ' ' && s[3])
         {
           s += 3;
           s2 = strchr (s, ' ');
           if ( s2 > s )
             {
-              parm->serialno = xtrymalloc ((s2 - s)+1);
-              if (parm->serialno)
+              *serialno = xtrymalloc ((s2 - s)+1);
+              if (*serialno)
                 {
-                  memcpy (parm->serialno, s, s2 - s);
-                  parm->serialno[s2 - s] = 0;
+                  memcpy (*serialno, s, s2 - s);
+                  (*serialno)[s2 - s] = 0;
                 }
             }
-          s -= 3;
         }
-
-      if (parm->fill_mode && *s0)
-        {
-          keyinfo_cache_item_t ci;
-          size_t n;
-
-          n = s? (s - s0) : strlen (s0);
-          ci = xtrymalloc (sizeof *ci + n);
-          if (!ci)
-            return gpg_error_from_syserror ();
-          memcpy (ci->hexgrip, s0, n);
-          ci->hexgrip[n] = 0;
-          ci->serialno = parm->serialno;
-          parm->serialno = NULL;
-          ci->next = parm->cache;
-          parm->cache = ci;
-        }
-
     }
   return 0;
 }
 
-
 /* Return the serial number for a secret key.  If the returned serial
- * number is NULL, the key is not stored on a smartcard.  Caller needs
- * to free R_SERIALNO.
- *
- * Take care: The cache is currently only used in the key listing and
- * it should not interfere with import or creation of new keys because
- * we assume that is done by another process.  However we assume that
- * in server mode the key listing is not directly followed by an import
- * and another key listing.
- */
+   number is NULL, the key is not stored on a smartcard.  Caller needs
+   to free R_SERIALNO.  */
 gpg_error_t
 gpgsm_agent_keyinfo (ctrl_t ctrl, const char *hexkeygrip, char **r_serialno)
 {
   gpg_error_t err;
   char line[ASSUAN_LINELENGTH];
-  keyinfo_cache_item_t ci;
-  struct keyinfo_status_parm_s parm = { NULL };
+  char *serialno = NULL;
 
   *r_serialno = NULL;
 
@@ -1460,73 +1289,20 @@ gpgsm_agent_keyinfo (ctrl_t ctrl, const char *hexkeygrip, char **r_serialno)
   if (!hexkeygrip || strlen (hexkeygrip) != 40)
     return gpg_error (GPG_ERR_INV_VALUE);
 
-  /* First try to fill the cache.  */
-  if ((opt.compat_flags & COMPAT_NO_KEYINFO_CACHE))
-    keyinfo_cache_disabled = 1;
+  snprintf (line, DIM(line), "KEYINFO %s", hexkeygrip);
 
-  if (!keyinfo_cache_disabled && !ctrl->keyinfo_cache_valid)
-    {
-      parm.fill_mode = 1;
-      err = assuan_transact (agent_ctx, "KEYINFO --list",
-                            NULL, NULL, NULL, NULL,
-                            keyinfo_status_cb, &parm);
-      if (err)
-        {
-          if (gpg_err_code (err) != GPG_ERR_FORBIDDEN)
-            log_error ("filling keyinfo cache failed: %s\n",
-                       gpg_strerror (err));
-          keyinfo_cache_disabled = 1;
-          release_a_keyinfo_cache (&parm.cache);
-          err = 0;  /* Fallback to single requests.  */
-        }
-      else
-        {
-          ctrl->keyinfo_cache_valid = 1;
-          ctrl->keyinfo_cache = parm.cache;
-          parm.cache = NULL;
-        }
-    }
-
-  /* Then consult the cache or send a query  */
-  if (ctrl->keyinfo_cache_valid)
-    {
-      for (ci = ctrl->keyinfo_cache; ci; ci = ci->next)
-        if (!strcmp (hexkeygrip, ci->hexgrip))
-          break;
-      if (ci)
-        {
-          xfree (parm.serialno);
-          parm.serialno = NULL;
-          err = 0;
-          if (ci->serialno)
-            {
-              parm.serialno = xtrystrdup (ci->serialno);
-              if (!parm.serialno)
-                err = gpg_error_from_syserror ();
-            }
-        }
-      else
-        err = gpg_error (GPG_ERR_NOT_FOUND);
-    }
-  else
-    {
-      snprintf (line, DIM(line), "KEYINFO %s", hexkeygrip);
-      parm.fill_mode = 0;
-      err = assuan_transact (agent_ctx, line, NULL, NULL, NULL, NULL,
-                             keyinfo_status_cb, &parm);
-    }
-
-  if (!err && parm.serialno)
+  err = assuan_transact (agent_ctx, line, NULL, NULL, NULL, NULL,
+                         keyinfo_status_cb, &serialno);
+  if (!err && serialno)
     {
       /* Sanity check for bad characters.  */
-      if (strpbrk (parm.serialno, ":\n\r"))
-        err = gpg_error (GPG_ERR_INV_VALUE);
+      if (strpbrk (serialno, ":\n\r"))
+        err = GPG_ERR_INV_VALUE;
     }
-
   if (err)
-    xfree (parm.serialno);
+    xfree (serialno);
   else
-    *r_serialno = parm.serialno;
+    *r_serialno = serialno;
   return err;
 }
 
